package nl.kallestruik.darena.arenas

import kotlinx.serialization.*
import kotlinx.serialization.json.Json
import nl.kallestruik.darena.types.ArenaFeature
import nl.kallestruik.darena.types.arena.*
import nl.kallestruik.darena.util.Logger
import nl.kallestruik.dlib.config.annotations.Description
import java.io.File

@Serializable
data class ArenaConfig(
    @Description(["The name of the arena."])
    var name: String,
    @Description(["The spawns that the arena will use. These alone are not enough to make a player spawn (See spawnRules)."])
    var spawns: MutableMap<String, ArenaSpawn> = mutableMapOf(),
    @Description(["The loadouts for the arena. These determine the items a player spawns with and what potion effects they get."])
    var loadouts: MutableMap<String, ArenaLoadout> = mutableMapOf(),
    @Description(["The checkpoints for the arena. Both normal checkpoints and finish points are defined here."])
    var checkpoints: MutableList<ArenaCheckpoint> = mutableListOf(),
    @Description([
        "The points that can be obtained by doing certain actions.",
        "These points are always awarded top down with the last repeating.",
        "For example when you set kill to 10, 5, and 0, the first player to get a kill will receive 10 points, the second 5, and every player after that will receive 0."])
    var points: ArenaPoints = ArenaPoints(),
    @Description([
            "The spawn rules used by the arena. These rules are used by many different components to link multiple spawns together with a set of items.",
            "There are two special spawn rules that are required for the operation of the plugin. These are:",
            "initial - Used by the plugin to do the initial round of spawning.",
            "spectator - Used by the plugin to spawn the spectators."])
    var spawnRules: MutableMap<String, ArenaSpawnRule> = mutableMapOf(),
    @Description(["The list of features that are enabled in the arena."])
    var features: MutableSet<ArenaFeature> = mutableSetOf(),
    @Description(["" +
            "The different borders used by the arena. These are defined using border script. Below a short explanation of the different keyword:",
            "CENTER <x> <z> - Set the center of the border to x,z. (ONLY AVAILABLE FOR VERTICAL BORDERS)",
            "SET <i> - On vertical borders set the with to i. On horizontal borders set the height to i.",
            "TRANSITION <i> <t> - Same as SET <i>, but over t seconds.",
            "WAIT <t> - Wait for t seconds."])
    var borders: ArenaBorders = ArenaBorders(),
    @SerialName("end-conditions")
    @Description(["The end conditions of the arena."])
    var endConditions: ArenaEndConditions = ArenaEndConditions(),
    @SerialName("spawn-pools")
    @Description(["The spawn pools that can be used by the arena. These are used to define groups of spawn rules that can be chosen at random by checkpoints."])
    var spawnPools: MutableMap<String, ArenaSpawnPool> = mutableMapOf(),
    @SerialName("break-allowed")
    @Description(["The blocks that players are allowed to break in the arena. Unless specified otherwise all blocks can be broken."])
    var breakAllowed: ArenaAllowList = ArenaAllowList(),
    @SerialName("place-allowed")
    @Description(["The blocks that players are allowed to place in the arena. Unless specified otherwise all blocks can be broken."])
    var placeAllowed: ArenaAllowList = ArenaAllowList(),
    @Description(["The different countdowns used by the arena."])
    var countdown: ArenaCountdown = ArenaCountdown(),
    @SerialName("block-remove-zone")
    @Description(["A area of blocks to remove at the start of the game."])
    var blockRemoveZone: ArenaBlockRemoveZone? = null,
) {
    @Transient
    lateinit var file: File

    fun save(toString: Boolean = false): String? {
        val json = Json {
            prettyPrint = true
        }.encodeToString(this)

        if (toString) {
            return json
        }

        file.writeText(json, Charsets.UTF_8)
        return null
    }

    companion object {
        fun load(file: File): ArenaConfig {
            Logger.trace(ArenaConfig::class, "load(file: ${file.path})")

            val config = Json.decodeFromString<ArenaConfig>(file.readText(Charsets.UTF_8))
            config.file = file

            return config
        }
    }
}
