package nl.kallestruik.darena.arenas.world

import nl.kallestruik.darena.exceptions.ArenaWorldCreationException
import nl.kallestruik.darena.exceptions.ArenaWorldLoadException
import nl.kallestruik.darena.exceptions.ArenaWorldSaveException
import nl.kallestruik.darena.managers.ConfigManager
import org.bukkit.Bukkit
import org.bukkit.Location
import org.bukkit.World
import org.bukkit.WorldCreator
import org.bukkit.plugin.java.JavaPlugin
import java.io.File
import java.nio.file.Files
import java.nio.file.Path
import kotlin.io.path.exists


class ArenaWorld(
    val name: String,
    val configManager: ConfigManager,
    val plugin: JavaPlugin,
) {
    lateinit var world: World
    val worldFolder = File(Bukkit.getServer().worldContainer, name)

    init {
        load()
        createWorld()
        world.isAutoSave = false
    }

    @Throws(ArenaWorldCreationException::class)
    private fun createWorld() {
        world = WorldCreator.name(name).generator(ArenaChunkGenerator()).createWorld()
            ?: throw ArenaWorldCreationException("Exception while creating bukkit world for arena \"$name\".")
    }

    @Throws(ArenaWorldLoadException::class)
    fun reset() {
        empty()
        Bukkit.unloadWorld(world, false)
        load()
        createWorld()
    }

    @Throws(ArenaWorldSaveException::class)
    fun save() {
        world.save()
        try {
            val savePath = Path.of(plugin.dataFolder.path, "worlds", name)

            if (savePath.exists())
                Files.walk(savePath).use { walk ->
                    walk.sorted(Comparator.reverseOrder()).forEach { path ->
                        Files.delete(path)
                    }
                }

            Files.walk(worldFolder.toPath()).use { walk ->
                walk.forEach { source ->
                    val destination = Path.of(
                        savePath.toString(), source.toString()
                            .substring(worldFolder.path.length)
                    )
                    // Don't copy uid.dat files because they stop bukkit from loading the world.
                    if (source.endsWith("uid.dat"))
                        return@forEach

                    Files.copy(source, destination)
                }
            }
        } catch (e: Exception) {
            throw ArenaWorldSaveException("There was an issue saving the world for arena \"$name\" to a safe location!", e)
        }
    }

    @Throws(ArenaWorldLoadException::class)
    private fun load() {
        try {
            val loadPath = Path.of(plugin.dataFolder.path, "worlds", name)
            if (!loadPath.exists())
                return

            Files.walk(worldFolder.toPath()).use { walk ->
                walk.sorted(Comparator.reverseOrder()).forEach { path ->
                    Files.delete(path)
                }
            }

            Files.walk(loadPath).use { walk ->
                walk.forEach { source ->
                    val destination = Path.of(
                        worldFolder.path, source.toString()
                            .substring(loadPath.toString().length)
                    )
                    // Don't copy uid.dat files because they stop bukkit from loading the world.
                    if (source.endsWith("uid.dat"))
                        return@forEach

                    Files.copy(source, destination)
                }
            }
        } catch (e: Exception) {
            throw ArenaWorldLoadException("There was an issue load the world for arena \"$name\"!", e)
        }
    }

    fun empty() {
        for (player in world.players) {
            player.spigot().respawn()
            player.teleport(Location(Bukkit.getWorld(configManager.spawnWorldName), configManager.spawnPos.x + 0.5, configManager.spawnPos.y + 0.5, configManager.spawnPos.z + 0.5))
        }
    }
}