package nl.kallestruik.darena.managers

import net.kyori.adventure.text.Component
import net.kyori.adventure.text.format.NamedTextColor
import nl.kallestruik.darena.arenas.Arena
import nl.kallestruik.darena.arenas.world.ArenaWorld
import nl.kallestruik.darena.exceptions.ArenaEditException
import nl.kallestruik.darena.types.EditSetting
import nl.kallestruik.darena.types.EditVisualization
import nl.kallestruik.darena.types.arena.ArenaSpawn
import nl.kallestruik.darena.types.arena.ArenaSpawnRule
import nl.kallestruik.darena.util.Logger
import nl.kallestruik.darena.util.RenderUtil
import nl.kallestruik.dlib.config.ConfigEntryInfo
import nl.kallestruik.dlib.config.ConfigHelper
import nl.kallestruik.dlib.gui.confirmationDialog
import nl.kallestruik.dlib.gui.textInputDialog
import org.bukkit.Color
import org.bukkit.GameMode
import org.bukkit.Location
import org.bukkit.Material
import org.bukkit.entity.EntityType
import org.bukkit.entity.Player
import org.bukkit.metadata.FixedMetadataValue
import org.bukkit.plugin.Plugin
import org.bukkit.scheduler.BukkitRunnable
import org.bukkit.scheduler.BukkitTask
import java.util.*

class EditManager(
    private val arenaManager: ArenaManager,
    private val plugin: Plugin
) {
    private var currentArena: Arena? = null
    private var currentWorld: ArenaWorld? = null
    private val currentSettings: MutableMap<UUID, MutableMap<EditSetting, Boolean>> = mutableMapOf()
    private val currentVisualizations: MutableMap<EditVisualization, Boolean> = mutableMapOf()
    private var checkpointVisualizationTask: BukkitTask? = null

    @Throws(ArenaEditException::class)
    fun load(arena: Arena) {
        Logger.trace(EditManager::class, "load(arena: ${arena.config.name})")

        if (isEditing())
            throw ArenaEditException("Cant start editing a world when another one is already loaded!")

        if (arenaManager.currentArena != null)
            throw ArenaEditException("Cant start editing a world when there is an arena in progress!")

        arena.world.reset()
        currentArena = arena
        currentWorld = arena.world
    }

    @Throws(ArenaEditException::class)
    fun save() {
        Logger.trace(EditManager::class, "save()")

        if (!isEditing())
            throw ArenaEditException("Cant save a world if you aren't editing one!")

        if (isVisualizationEnabled(EditVisualization.SPAWNS)) {
            removeSpawnVisualizers()
        }

        currentWorld!!.save()
        saveConfig()

        if (isVisualizationEnabled(EditVisualization.SPAWNS)) {
            addSpawnVisualizers()
        }
    }

    @Throws(ArenaEditException::class)
    fun tp(player: Player) {
        Logger.trace(EditManager::class, "tp(player: ${player.name})")

        if (!isEditing())
            throw ArenaEditException("Cant teleport to a world if none is being edited!")

        player.teleport(Location(currentWorld!!.world, 0.5, 100.0, 0.5))
        player.gameMode = GameMode.CREATIVE
    }

    fun finish() {
        Logger.trace(EditManager::class, "finish()")

        if (!isEditing())
            throw ArenaEditException("Cant stop editing what is not being edited!")

        save()
        currentWorld!!.empty()
        currentArena = null
        currentWorld = null
        currentSettings.clear()
        checkpointVisualizationTask?.cancel()
    }

    fun isEditing(): Boolean {
        return currentWorld != null
    }

//    fun getConfigValue(path: String): ConfigEntryInfo {
//        Logger.trace(EditManager::class, "getConfigValue(path: $path)")
//
//        if (!isEditing())
//            throw ArenaEditException("You arent editing anything!")
//
//        return ConfigHelper.getInfoByPath(path, currentArena!!.config)
//    }

    fun toggleFeature(player: Player, setting: EditSetting): Boolean {
        val settingsMap = currentSettings.getOrPut(player.uniqueId, ::mutableMapOf)
        val newValue = !(settingsMap[setting] ?: false)
        settingsMap[setting] = newValue
        return newValue
    }

    fun saveConfig() {
        if (!isEditing()) {
            throw ArenaEditException("You cant save something that has not been changed!")
        }

        arenaManager.saveConfig(currentArena!!)
    }

    fun shouldProcessEvent(player: Player?, setting: EditSetting): Boolean {
        if (currentArena == null)
            return false

        if (!isEditing())
            return false

        if (player == null)
            return false

        if (setting != EditSetting.ALWAYS && !playerHasFeatureEnabled(player, setting))
            return false

        return true
    }

    private fun playerHasFeatureEnabled(player: Player, setting: EditSetting): Boolean {
        return currentSettings[player.uniqueId]?.get(setting) ?: false
    }

    fun addSpawn(name: String, spawn: ArenaSpawn): Boolean {
        Logger.trace(EditManager::class, "finishSpawn(name: $name, spawn: $spawn)")

        if (!isEditing())
            return false

        if (name in currentArena!!.config.spawns)
            return false

        currentArena!!.config.spawns[name] = spawn
        refreshSpawnVisualizer()
        return true
    }

    fun isVisualizationEnabled(visualization: EditVisualization): Boolean {
        return currentVisualizations[visualization] ?: false
    }

    fun toggleSpawnVisibility() {
        if (!isEditing()) {
            throw ArenaEditException("You cant visualize what isn't there!")
        }

        val newValue = !isVisualizationEnabled(EditVisualization.SPAWNS)
        currentVisualizations[EditVisualization.SPAWNS] = newValue

        if (newValue) {
            addSpawnVisualizers()
        } else {
            removeSpawnVisualizers()
        }

    }

    private fun addSpawnVisualizers() {
        val world = currentWorld!!.world
        for (entry in currentArena!!.config.spawns) {
            val entity = world.spawnEntity(entry.value.toLocation(world), EntityType.ARMOR_STAND)
            entity.setGravity(false)
            entity.isCustomNameVisible = true
            entity.customName(Component.text("SPAWN: ${entry.key}"))
            entity.setMetadata("ArenaSpawn", FixedMetadataValue(plugin, entry.key))
        }
    }

    private fun removeSpawnVisualizers() {
        for (entity in currentWorld!!.world.entities) {
            if (!entity.hasMetadata("ArenaSpawn"))
                continue

            entity.remove()
        }
    }

    fun refreshSpawnVisualizer() {
        if (!isEditing())
            return

        if (!isVisualizationEnabled(EditVisualization.SPAWNS))
            return

        removeSpawnVisualizers()
        addSpawnVisualizers()
    }

    fun removeSpawn(spawnName: String) {
        if (!isEditing()) {
            throw ArenaEditException("Cant remove a spawn if you arent editing an arena!")
        }

        currentArena!!.config.spawns.remove(spawnName)
        refreshSpawnVisualizer()
    }

    fun toggleCheckpointVisibility() {
        if (!isEditing()) {
            throw ArenaEditException("You cant visualize what isn't there!")
        }

        val newValue = !isVisualizationEnabled(EditVisualization.CHECKPOINTS)
        currentVisualizations[EditVisualization.CHECKPOINTS] = newValue
        if (newValue) {
            checkpointVisualizationTask = CheckpointVisualizationRunnable(currentArena!!).runTaskTimer(plugin, 0, 5)
        } else {
            checkpointVisualizationTask?.cancel()
        }
    }

    fun addSpawnToRuleGUI(player: Player, spawnName: String) {
        if (!isEditing())
            throw ArenaEditException("You arent editing an arena!")

        if (spawnName !in currentArena!!.config.spawns)
            throw ArenaEditException("That spawn does not exist!")

        textInputDialog(player, plugin) {
            prompt = Component.text("Rule name")
            item {
                material = Material.LIME_DYE
            }

            onCancel {
                player.sendMessage(Component.text("Canceled adding $spawnName to a rule."))
            }

            onFinish { name ->
                val rule: ArenaSpawnRule
                if (name in currentArena!!.config.spawnRules) {
                    rule = currentArena!!.config.spawnRules[name]!!

                    player.sendMessage(Component.text("Added $spawnName to rule $name."))
                } else {
                    rule = ArenaSpawnRule()

                    player.sendMessage(Component.text("Created new rule $name and added $spawnName to it."))
                }

                rule.spawns.add(spawnName)
                currentArena!!.config.spawnRules[name] = rule
            }
        }
    }

    fun getSpawn(spawnName: String): ArenaSpawn {
        if (!isEditing())
            throw ArenaEditException("You cant get a spawn when you arent editing an arena!")

        return currentArena!!.config.spawns[spawnName]
            ?: throw ArenaEditException("That spawn does not exist!")
    }

    fun removeSpawnGUI(player: Player, spawnName: String) {
        confirmationDialog(player, plugin) {
            title = Component.text("Remove spawn?")
            description = Component
                .text("Are you sure that you want to remove this spawn?")
                .color(NamedTextColor.RED)

            onConfirm {
                removeSpawn(spawnName)
                refreshSpawnVisualizer()
                player.sendMessage(Component.text("Spawn $spawnName removed!"))
            }
        }
    }

    private class CheckpointVisualizationRunnable(
        val arena: Arena,
    ): BukkitRunnable() {
        override fun run() {
            for (checkpoint in arena.config.checkpoints) {
                val lower = checkpoint.lower
                val upper = checkpoint.upper

                RenderUtil.drawAABB(
                    arena.world.world,
                    lower.x, lower.y, lower.z,
                    upper.x, upper.y, upper.z,
                    Color.AQUA
                )
            }
        }

    }
}