package nl.kallestruik.darena.managers

import net.kyori.adventure.text.Component
import net.kyori.adventure.text.event.HoverEvent
import nl.kallestruik.darena.exceptions.TeamExistsException
import nl.kallestruik.darena.exceptions.TeamNotFoundException
import nl.kallestruik.darena.types.Team
import nl.kallestruik.darena.util.ConfigHelper
import nl.kallestruik.darena.util.Logger
import org.bukkit.Bukkit
import org.bukkit.configuration.file.YamlConfiguration
import java.io.File
import java.util.*

class TeamManager(
    val teamsFile: File
) {
    private val teams: MutableMap<String, Team> = HashMap()
    private val teamMembers: MutableMap<Team, MutableList<UUID>> = HashMap()
    private val teamForPlayer: MutableMap<UUID, Team> = HashMap()
    var autoTeams: Boolean = true

    fun load() {
        // Make sure all lists are empty so we don't have leftovers.
        teams.clear()
        teamMembers.clear()
        teamForPlayer.clear()

        val config = ConfigHelper.getOrCreateConfig(teamsFile, "template/teams.yml")
        autoTeams = config.getBoolean("auto-teams")
        if (config.isConfigurationSection("teams")) {
            for (teamName in config.getConfigurationSection("teams")!!.getKeys(false)) {
                Logger.debug(TeamManager::class, "Loading team $teamName")
                val team = Team.fromConfig(config.getConfigurationSection("teams.$teamName")!!)
                teams[teamName] = team
            }
        }

        if (config.contains("members")) {
            for (memberString in config.getStringList("members")) {
                val parts = memberString.split(":")
                val uuid = UUID.fromString(parts[0])
                val teamName = parts[1]
                addPlayerToTeam(teamName, uuid)
            }
        }
    }

    fun save() {
        val config = YamlConfiguration()
        config.set("auto-teams", autoTeams)

        val teamSection = config.createSection("teams")
        for (team in teams.values) {
            team.save(teamSection)
        }

        config.set("members", teamForPlayer.map {
            "${it.key}:${it.value.name}"
        })

        config.save(teamsFile)
    }

    @Throws(TeamExistsException::class)
    fun createTeam(teamName: String): Team {
        if (teams.containsKey(teamName))
            throw TeamExistsException("The team $teamName already exists!")

        val team = Team(teamName)
        teams[teamName] = team
        return team
    }

    fun removeTeam(team: Team) {
        teams.remove(team.name)
        for (entry in teamForPlayer) {
            if (entry.value == team) {
                removePlayerFromTeam(entry.key)
            }
        }
        team.mcTeam.unregister()
    }

    @Throws(TeamNotFoundException::class)
    fun addPlayerToTeam(teamName: String, uuid: UUID) {
        Logger.trace(TeamManager::class, "addPlayerToTeam(teamName: $teamName, uuid: $uuid)")
        val team = getTeamByName(teamName)

        addPlayerToTeam(team, uuid)
    }

    fun addPlayerToTeam(team: Team, uuid: UUID) {
        Logger.trace(TeamManager::class, "addPlayerToTeam(team: ${team.name}, uuid: $uuid)")
        teamMembers.getOrPut(team) { ArrayList() }.add(uuid)
        teamForPlayer[uuid] = team

        val player = Bukkit.getPlayer(uuid)
        if (player != null)
            team.mcTeam.addEntry(player.name)
    }

    @Throws(TeamNotFoundException::class)
    fun getTeamForPlayer(uuid: UUID): Team? {
        return teamForPlayer[uuid]
    }

    @Throws(TeamNotFoundException::class)
    fun getTeamByName(teamName: String): Team {
        return teams[teamName] ?: throw TeamNotFoundException("Team with name $teamName does not exist!")
    }

    @Throws(TeamNotFoundException::class)
    fun getTeamMembers(teamName: String): List<UUID> {
        val team = getTeamByName(teamName)
        return getTeamMembers(team)
    }

    fun getTeamMembers(team: Team): List<UUID> {
        return teamMembers[team] ?: return emptyList()
    }

    fun removeAllTeams(): Int {
        val count = teams.size
        for (team in getAllTeams()) {
            removeTeam(team)
        }
        return count
    }

    fun autoTeam() {
        for (player in Bukkit.getOnlinePlayers()) {
            val team = createTeam(player.name)
            addPlayerToTeam(team, player.uniqueId)
        }
    }

    fun removePlayerFromTeam(uuid: UUID) {
        val team = getTeamForPlayer(uuid)
        teamMembers[team]?.remove(uuid)
        teamForPlayer.remove(uuid)
        team?.mcTeam?.removeEntry(Bukkit.getPlayer(uuid)!!.name)
    }

    fun getAllTeamNames() = teams.keys

    fun getAllTeams() = teams.values

    fun getMemberTooltip(team: Team): HoverEvent<Component> {
        var hoverComponent = Component.text("Members:")
        getTeamMembers(team).forEach {
            hoverComponent = hoverComponent.append(Component.text("\n${Bukkit.getOfflinePlayer(it).name}"))
        }
        return hoverComponent.asHoverEvent()
    }

    fun getTeamComponent(team: Team): Component {
        return Component
            .text(team.name)
            .color(team.color)
            .hoverEvent(getMemberTooltip(team))
    }
}